"""Convert ADIF ADI content to dictionary and vice versa"""

import re
import datetime
from collections.abc import Iterator

from adif_file.__version__ import __version__ as __version_str__

__proj_name__ = 'PyADIF-File'
__version__ = __version_str__[1:].split('-')[0]


class TooMuchHeadersException(Exception):
    pass


class TagDefinitionException(Exception):
    pass


class StringNotASCIIException(Exception):
    pass


class IllegalParameterException(Exception):
    pass


class IllegalDataTypeException(Exception):
    pass


REGEX_ASCII = re.compile(r'[ -~\n\r]*')
REGEX_PARAM = re.compile(r'[a-zA-Z][a-zA-Z_0-9]*')


def unpack(data: str) -> dict:
    """Unpack header or record part to dictionary
    The parameters are converted to uppercase"""

    unpacked = {}

    start = -1
    end = -1
    length = 0

    while start < len(data):
        try:
            start = data.index('<', end + 1 + length)
        except ValueError:
            break

        end = data.index('>', start)
        tag = data[start + 1:end]
        dtype = None
        try:
            tag_def = tag.split(':')
            param = tag_def[0]
            length = tag_def[1]
            if len(tag_def) == 3:
                dtype = tag_def[2]
        except ValueError:
            raise TagDefinitionException('Wrong tag definition')

        try:
            length = int(length)
        except ValueError:
            raise TagDefinitionException('Wrong length')

        value = data[end + 1:end + 1 + length]
        if param.upper().startswith('USERDEF'):
            if 'USERDEFS' not in unpacked:
                unpacked['USERDEFS'] = []
            unpacked['USERDEFS'].append({'dtype': dtype,
                                         'userdef': value})
        else:
            unpacked[param.upper()] = value

    return unpacked


def loadi(adi: str, skip: int = 0) -> Iterator[dict]:
    """Turn ADI formated string to header/records as an iterator over dict
    The skip option is useful if you want to watch a file for new records only. This saves processing time.

    :param adi: the ADI data
    :param skip: skip first number of records (does not apply for header)
    :return: an iterator of records (first record is the header even if not available)
    """

    record_data = adi
    if not adi.startswith('<'):  # If a header is available
        hr_list = re.split(r'<[eE][oO][hH]>', adi)
        if len(hr_list) > 2:
            raise TooMuchHeadersException()

        yield unpack(hr_list[0])
        record_data = hr_list[1]
    else:  # Empty record for missing header
        yield {}

    i = 0
    for rec in re.finditer(r'(.*?)<[eE][oO][rR]>', record_data, re.S):
        if i >= skip:
            yield unpack(rec.groups()[0])
        i += 1


def loads(adi: str, skip: int = 0) -> dict:
    """Turn ADI formated string to dictionary
    The parameters are converted to uppercase

        {
        'HEADER': None,
        'RECORDS': [list of records]
        }

    The skip option is useful if you want to watch a file for new records only. This saves processing time.
    In this case consider to use loadi() directly.

    :param adi: the ADI data
    :param skip: skip first number of records (does not apply for header)
    :return: the ADI as a dict
    """

    doc = {'HEADER': None,
           'RECORDS': []
           }

    first = True
    for rec in loadi(adi, skip):
        if first:
            doc['HEADER'] = rec
            first = False
        else:
            doc['RECORDS'].append(rec)

    return doc


def load(file_name: str, skip: int = 0, encoding=None) -> dict:
    """Load ADI formated file to dictionary
       The parameters are converted to uppercase

           {
           'HEADER': None,
           'RECORDS': [list of records]
           }

       The skip option is useful if you want to watch a file for new records only. This saves processing time.
       In this case consider to use loadi() directly.

       :param file_name: the file name where the ADI data is stored
       :param skip: skip first number of records (does not apply for header)
       :param encoding: the file encoding
       :return: the ADI as a dict
       """

    with open(file_name, encoding=encoding) as af:
        data = af.read()

    return loads(data, skip)


def pack(param: str, value: str, dtype: str = None) -> str:
    """Generates ADI tag if value is not empty
    Does not generate tags for *_INTL types as required by specification.

    :param param: the tag parameter (converte to uppercase)
    :param value: the tag value (or tag definition if param is a USERDEF field)
    :param dtype: the optional datatype (mainly used for USERDEFx in header)
    :return: <param:length>value
    """

    if not re.fullmatch(REGEX_PARAM, param):
        raise IllegalParameterException(f'Parameter "{param}" contains not allowed characters')

    if not param.upper().endswith('_INTL'):
        if isinstance(value, str) and not re.fullmatch(REGEX_ASCII, value):
            raise StringNotASCIIException(f'Value "{value}" in parameter "{param}" contains non ASCII characters')

        if dtype:
            if len(dtype) > 1 or dtype not in 'BNDTSEL':
                raise IllegalDataTypeException(f'Datatype "{dtype}" in "{param}"')
            return f'<{param.upper()}:{len(str(value))}:{dtype}>{value}' if value else ''
        else:
            return f'<{param.upper()}:{len(str(value))}>{value}' if value else ''
    else:
        return ''


def dumpi(data_dict: dict, comment: str = 'ADIF export by ' + __proj_name__,
          linebreaks: bool = True, spaces: int = 1) -> Iterator[str]:
    """Takes a dictionary and converts it to ADI format
    Parameters can be in upper or lower case. The output is upper case. The user must take care
    that parameters are not doubled!
    *_INTL parameters are ignored as they are not allowed in ADI.
    Empty records are skipped.

    If 'HEADER' is present the comment is added and missing header fields are filled with defaults.
    The header can contain a list of user definitions as USERDEFS. Each user definition is expected as a dictionary
    with datatype as "dtype" and field definition as "userdef" instead of a string value.

    :param data_dict: the dictionary with header and records
    :param comment: the comment to induce the header
    :param linebreaks: Format output with additional linebreaks for readability
    :param spaces: Number of spaces between fields
    :return: an iterator of chunks of the ADI"""

    default = {'ADIF_VER': '3.1.4',
               'PROGRAMID': __proj_name__,
               'PROGRAMVERSION': __version__,
               'CREATED_TIMESTAMP': datetime.datetime.utcnow().strftime('%Y%m%d %H%M%S')
               }

    field_separator = ' ' * spaces if spaces >= 0 else ' '

    if 'HEADER' in data_dict:
        data = comment + ' \n'

        for p in data_dict['HEADER']:
            if p.upper() in ('ADIF_VER', 'PROGRAMID', 'PROGRAMVERSION', 'CREATED_TIMESTAMP'):
                data += pack(p.upper(), data_dict['HEADER'][p]) + ('\n' if linebreaks else field_separator)
                default.pop(p.upper())
            elif p.upper() == 'USERDEFS':
                for i, u in enumerate(data_dict['HEADER'][p], 1):
                    data += pack(f'USERDEF{i}', u['userdef'], u['dtype']) + ('\n' if linebreaks else field_separator)
        for p in default:
            data += pack(p, default[p]) + ('\n' if linebreaks else field_separator)
        data += '<EOH>'
        yield data

    if 'RECORDS' in data_dict:
        for r in data_dict['RECORDS']:
            data = ''
            empty = True
            for i, pv in enumerate(zip(r.keys(), r.values()), 1):
                tag = pack(pv[0].upper(), pv[1])
                if tag:
                    empty = False
                    if linebreaks:
                        data += tag + ('\n' if i % 5 == 0 else field_separator)
                    else:
                        data += tag + field_separator
            if not data.endswith('\n'):
                data += '\n' if linebreaks else ''

            if not empty:
                data += '<EOR>'
                yield data


def dumps(data_dict: dict, comment: str = 'ADIF export by ' + __proj_name__, linebreaks: bool = True, **params) -> str:
    """Takes a dictionary and converts it to ADI format
    Parameters can be in upper or lower case. The output is upper case. The user must take care
    that parameters are not doubled!
    *_INTL parameters are ignored as they are not allowed in ADI.
    Empty records are skipped.

    If 'HEADER' is present the comment is added and missing header fields are filled with defaults.
    The header can contain a list of user definitions as USERDEFS. Each user definition is expected as a dictionary
    with datatype as "dtype" and field definition as "userdef" instead of a string value.

    :param data_dict: the dictionary with header and records
    :param comment: the comment to induce the header
    :param linebreaks: Format output with additional linebreaks for readability
    :return: the complete ADI as a string"""

    line_separator = '\n\n' if linebreaks else '\n'

    return line_separator.join(list(dumpi(data_dict, comment, linebreaks=linebreaks, **params)))


def dump(file_name: str, data_dict: dict, comment: str = 'ADIF export by ' + __proj_name__,
         linebreaks: bool = True, encoding='ascii', **params):
    """Takes a dictionary and stores it to filename in ADI format
    Parameters can be in upper or lower case. The output is upper case. The user must take care
    that parameters are not doubled!
    *_INTL parameters are ignored as they are not allowed in ADI.
    Empty records are skipped.

    If 'HEADER' is present the comment is added and missing header fields are filled with defaults.
    The header can contain a list of user definitions as USERDEFS. Each user definition is expected as a dictionary
    with datatype as "dtype" and field definition as "userdef" instead of a string value.

    :param file_name: the filename to store the ADI data to
    :param data_dict: the dictionary with header and records
    :param comment: the comment to induce the header
    :param linebreaks: format output with additional linebreaks for readability
    :param encoding: the file encoding
    :return: the complete ADI as a string"""

    with open(file_name, 'w', encoding=encoding) as af:
        first = True
        for chunk in dumpi(data_dict, comment, linebreaks=linebreaks, **params):
            if first:
                first = False
            else:
                af.write('\n\n' if linebreaks else '\n')

            af.write(chunk)


__all__ = ['load', 'loads', 'loadi', 'dump', 'dumps', 'dumpi',
           'TooMuchHeadersException', 'TagDefinitionException',
           'IllegalDataTypeException', 'IllegalParameterException',
           'StringNotASCIIException']



import os
import unittest

import adif_file.adx


def get_file_path(file):
    return os.path.join(os.path.dirname(__file__), file)


class LoadADX(unittest.TestCase):
    def test_10_goodfile(self):
        adx_exp_dict = {
            'HEADER': {'ADIF_VER': '3.1.4',
                       'CREATED_TIMESTAMP': '20231204 100000',
                       'PROGRAMID': 'PyADIF-File',
                       'PROGRAMVERSION': '1'},
            'RECORDS': [{'CALL': 'XX1XXX',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1100',
                         'QTH': 'Test'
                         },
                        {'CALL': 'YY1YYY',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1200',
                         'QTH_INTL': 'Töst',
                         'APP':
                             {
                                 '@PROGRAMID': 'TESTAPP',
                                 '@FIELDNAME': 'TESTFIELD',
                                 '@TYPE': 'I',
                                 '$': 'Test',
                             },
                         }]
        }

        self.maxDiff = None
        self.assertDictEqual(adx_exp_dict, adif_file.adx.load(get_file_path('testdata/goodfile.adx')))

    def test_20_badfile(self):
        self.assertRaises(adif_file.adx.XmlSyntaxError, adif_file.adx.load,
                          get_file_path('testdata/goodfile.txt'))

    def test_20_badxml(self):
        self.assertRaises(adif_file.adx.MalformedValueException, adif_file.adx.load,
                          get_file_path('testdata/badfile1.adx'), True)
        self.assertRaises(adif_file.adx.UndefinedElementException, adif_file.adx.load,
                          get_file_path('testdata/badfile2.adx'), True)


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import adif_file.adi


def get_file_path(file):
    return os.path.join(os.path.dirname(__file__), file)


class DumpADI(unittest.TestCase):
    def test_10_pack_header_tag(self):
        self.assertEqual('<PROGRAMID:8>Testprog', adif_file.adi.pack('PROGRAMID', 'Testprog'))
        self.assertEqual('<USERDEF1:4:N>Test', adif_file.adi.pack('USERDEF1', 'Test', 'N'))
        self.assertEqual('<USERDEF1:19:E>SweaterSize,{S,M,L}',
                         adif_file.adi.pack('USERDEF1', 'SweaterSize,{S,M,L}', 'E'))

        self.assertRaises(adif_file.adi.IllegalDataTypeException, adif_file.adi.pack, 'USERDEF1', 'SweaterSize,{S,M,L}', 'X')
        self.assertRaises(adif_file.adi.IllegalDataTypeException, adif_file.adi.pack, 'USERDEF1', 'SweaterSize,{S,M,L}', 'NN')

    def test_15_pack_record_tag(self):
        self.assertEqual('<NAME:5>Joerg', adif_file.adi.pack('NAME', 'Joerg'))
        self.assertEqual('<NAME:5>Joerg', adif_file.adi.pack('name', 'Joerg'))
        self.assertEqual('<NAME:5>Joerg', adif_file.adi.pack('Name', 'Joerg'))
        self.assertEqual('', adif_file.adi.pack('name_intl', 'Joerg'))
        self.assertEqual('<APP_TESTAPP_CHANNEL:2:N>24', adif_file.adi.pack('APP_TESTAPP_CHANNEL', 24, dtype='N'))

        self.assertEqual('<MY_NAME:5>Peter', adif_file.adi.pack('MY_Name', 'Peter'))

        self.assertRaises(adif_file.adi.StringNotASCIIException, adif_file.adi.pack, 'NAME', 'Jörg')
        self.assertRaises(adif_file.adi.StringNotASCIIException, adif_file.adi.pack, 'NAME', 'Schloß')
        self.assertRaises(adif_file.adi.IllegalParameterException, adif_file.adi.pack, 'MY_ NAME', 'Peter')
        self.assertRaises(adif_file.adi.IllegalParameterException, adif_file.adi.pack, 'MY~NAME', 'Peter')

        # noinspection PyTypeChecker
        self.assertEqual('<DIST:2>99', adif_file.adi.pack('DIST', 99))
        # noinspection PyTypeChecker
        self.assertEqual('<FREQ:5>0.138', adif_file.adi.pack('freq', 0.138))

    def test_20_dump_header(self):
        adi_dict = {
            'HEADER': {'PROGRAMID': 'TProg',
                       'ADIF_VER': '3',
                       'PROGRAMVERSION': '1',
                       'CREATED_TIMESTAMP': '1234'},
        }

        exp_hdr = '''ADIF export by PyADIF-File 
<PROGRAMID:5>TProg
<ADIF_VER:1>3
<PROGRAMVERSION:1>1
<CREATED_TIMESTAMP:4>1234
<EOH>'''

        self.assertEqual(exp_hdr, adif_file.adi.dumps(adi_dict))

        # Test same with udef
        adi_udef = [{'dtype': 'E',
                     'userdef': 'Test,{A,B,C}'},
                    {'dtype': 'N',
                     'userdef': 'Test2,{5:20}'}]

        exp_hdr_udef = '''ADIF export by PyADIF-File 
<PROGRAMID:5>TProg
<ADIF_VER:1>3
<PROGRAMVERSION:1>1
<CREATED_TIMESTAMP:4>1234
<USERDEF1:12:E>Test,{A,B,C}
<USERDEF2:12:N>Test2,{5:20}
<EOH>'''

        adi_dict['HEADER']['USERDEFS'] = adi_udef
        self.assertEqual(exp_hdr_udef, adif_file.adi.dumps(adi_dict))

    def test_25_dump_records(self):
        adi_dict = {
            'RECORDS': [{'TEST1': 'test',
                         'TEST2': 'test2'},
                        {'TEST1': 'test3',
                         'TEST2': 'test4\r\ntest5'}]}

        adi_exp = '''<TEST1:4>test <TEST2:5>test2 
<EOR>

<TEST1:5>test3 <TEST2:12>test4\r\ntest5 
<EOR>'''

        self.assertEqual(adi_exp, adif_file.adi.dumps(adi_dict))

    def test_30_dump_a_file(self):
        adi_dict = {
            'HEADER': {'PROGRAMID': 'TProg',
                       'ADIF_VER': '3',
                       'PROGRAMVERSION': '1',
                       'CREATED_TIMESTAMP': '1234'},
            'RECORDS': [{'TEST1': 'test',
                         'TEST2': 'test2'},
                        {'TEST1': 'test3',
                         'TEST2': 'test4'}]
        }

        adi_exp = '''ADIF export by PyADIF-File 
<PROGRAMID:5>TProg
<ADIF_VER:1>3
<PROGRAMVERSION:1>1
<CREATED_TIMESTAMP:4>1234
<EOH>

<TEST1:4>test <TEST2:5>test2 
<EOR>

<TEST1:5>test3 <TEST2:5>test4 
<EOR>'''

        temp_file = get_file_path('testdata/~test.adi')

        adif_file.adi.dump(temp_file, adi_dict)

        self.assertTrue(os.path.isfile(temp_file))

        with open(temp_file) as af:
            self.assertEqual(adi_exp, af.read())

        os.remove(temp_file)

    def test_31_dump_a_file_ln_sp(self):
            adi_dict = {
                'HEADER': {'PROGRAMID': 'TProg',
                           'ADIF_VER': '3',
                           'PROGRAMVERSION': '1',
                           'CREATED_TIMESTAMP': '1234'},
                'RECORDS': [{'TEST1': 'test',
                             'TEST2': 'test2'},
                            {'TEST1': 'test3',
                             'TEST2': 'test4'}]
            }

            adi_exp = '''ADIF export by PyADIF-File 
<PROGRAMID:5>TProg  <ADIF_VER:1>3  <PROGRAMVERSION:1>1  <CREATED_TIMESTAMP:4>1234  <EOH>
<TEST1:4>test  <TEST2:5>test2  <EOR>
<TEST1:5>test3  <TEST2:5>test4  <EOR>'''

            temp_file = get_file_path('testdata/~test.adi')

            adif_file.adi.dump(temp_file, adi_dict, linebreaks=False, spaces=2)

            self.assertTrue(os.path.isfile(temp_file))

            with open(temp_file) as af:
                self.assertEqual(adi_exp, af.read())

            os.remove(temp_file)

    def test_40_dump_no_change(self):
        adi_dict = {
            'HEADER': {'PROGRAMID': 'TProg',
                       'PROGRAMVERSION': '1',
                       'CREATED_TIMESTAMP': '1234'},
            'RECORDS': [{'TEST1': 'test',
                         'TEST2': 'test2'},
                        {'TEST1': 'test3',
                         'TEST2': 'test4'}]
        }
        adi_dict_sav = adi_dict.copy()

        adi_exp = '''ADIF export by PyADIF-File 
<PROGRAMID:5>TProg
<PROGRAMVERSION:1>1
<CREATED_TIMESTAMP:4>1234
<ADIF_VER:5>3.1.4
<EOH>

<TEST1:4>test <TEST2:5>test2 
<EOR>

<TEST1:5>test3 <TEST2:5>test4 
<EOR>'''

        temp_file = get_file_path('testdata/~test.adi')

        adif_file.adi.dump(temp_file, adi_dict)
        self.assertDictEqual(adi_dict_sav, adi_dict)

        self.assertTrue(os.path.isfile(temp_file))

        with open(temp_file) as af:
            self.assertEqual(adi_exp, af.read())

        os.remove(temp_file)


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import adif_file.adi


def get_file_path(file: str):
    return os.path.join(os.path.dirname(__file__), file)


class LoadADI(unittest.TestCase):
    def test_10_unpack_header(self):
        adi_hdr1 = '''ADIF Export by Testprog
<ADIF_VER:5>3.1.4 <PROGRAMID:8>Testprog <PROGRAMVERSION:4>v0.2'''
        adi_hdr2 = '''ADIF Export by Testprog
    <ADIF_VER:5>3.1.4
    <PROGRAMID:8>Testprog 
    <PROGRAMVERSION:4>v0.2'''
        exp_dict = {'ADIF_VER': '3.1.4', 'PROGRAMID': 'Testprog', 'PROGRAMVERSION': 'v0.2'}

        self.assertDictEqual(exp_dict, adif_file.adi.unpack(adi_hdr1))
        self.assertDictEqual(exp_dict, adif_file.adi.unpack(adi_hdr2))

    def test_15_unpack_userdef(self):
        adi_hdr1 = '''ADIF Export by Testprog
<ADIF_VER:5>3.1.4 <PROGRAMID:8>Testprog <PROGRAMVERSION:4>v0.2 <USERDEF1:4:N>Test'''
        adi_hdr2 = '''ADIF Export by Testprog
<ADIF_VER:5>3.1.4hh <PROGRAMID:8>Testprog <PROGRAMVERSION:4>v0.2 <USERDEF1:4:G>Test <USERDEF2:5:L>TestX'''

        exp_dict1 = {'ADIF_VER': '3.1.4', 'PROGRAMID': 'Testprog', 'PROGRAMVERSION': 'v0.2',
                     'USERDEFS': [{'dtype': 'N',
                                   'userdef': 'Test'}]}
        exp_dict2 = {'ADIF_VER': '3.1.4', 'PROGRAMID': 'Testprog', 'PROGRAMVERSION': 'v0.2',
                     'USERDEFS': [{'dtype': 'G',
                                   'userdef': 'Test'},
                                  {'dtype': 'L',
                                   'userdef': 'TestX'}]}

        self.assertDictEqual(exp_dict1, adif_file.adi.unpack(adi_hdr1))
        self.assertDictEqual(exp_dict2, adif_file.adi.unpack(adi_hdr2))

    def test_20_unpack_record(self):
        adi_rec_app = '<APP_TESTAPP_CHANNEL:2:N>24'
        adi_rec_name = '<NAME:4>Test'

        self.assertDictEqual({'APP_TESTAPP_CHANNEL': '24'}, adif_file.adi.unpack(adi_rec_app))
        self.assertDictEqual({'NAME': 'Test'}, adif_file.adi.unpack(adi_rec_name))

    def test_50_goodfile(self):
        adi_dict = adif_file.adi.load(get_file_path('testdata/goodfile.txt'))

        self.assertIn('HEADER', adi_dict)
        self.assertIn('RECORDS', adi_dict)
        self.assertEqual(3, len(adi_dict['HEADER']))
        self.assertEqual(5, len(adi_dict['RECORDS']))

    def test_52_no_header(self):
        adi_dict = adif_file.adi.load(get_file_path('testdata/goodfile_no_h.txt'))

        self.assertIn('HEADER', adi_dict)
        self.assertIn('RECORDS', adi_dict)
        self.assertEqual(0, len(adi_dict['HEADER']))
        self.assertEqual(5, len(adi_dict['RECORDS']))

    def test_55_toomuchheaders(self):
        self.assertRaises(adif_file.adi.TooMuchHeadersException, adif_file.adi.load,
                          get_file_path('testdata/toomuchheadersfile.txt'))

    def test_60_skiprecords(self):
        adi_dict = adif_file.adi.load(get_file_path('testdata/goodfile.txt'), 3)
        rec_dict = {'QSO_DATE': '20231008', 'TIME_ON': '1754', 'RST_SENT': '59',
                    'RST_RCVD': '59', 'BAND': '2190M', 'MODE': 'AM', 'FREQ': '0.137',
                    'TX_PWR': '4.0', 'STATION_CALLSIGN': 'XX1XXX', 'MY_GRIDSQUARE': 'JO35uj27'}

        self.assertIn('HEADER', adi_dict)
        self.assertIn('RECORDS', adi_dict)
        self.assertEqual(3, len(adi_dict['HEADER']))
        self.assertEqual(2, len(adi_dict['RECORDS']))
        self.assertDictEqual(rec_dict, adi_dict['RECORDS'][0])

    def test_65_skiprec_noh(self):
        adi_dict = adif_file.adi.load(get_file_path('testdata/goodfile_no_h.txt'), 3)
        rec_dict = {'QSO_DATE': '20231008', 'TIME_ON': '1754', 'RST_SENT': '59',
                    'RST_RCVD': '59', 'BAND': '2190M', 'MODE': 'AM', 'FREQ': '0.137',
                    'TX_PWR': '4.0', 'STATION_CALLSIGN': 'XX1XXX', 'MY_GRIDSQUARE': 'JO35uj27'}

        self.assertIn('HEADER', adi_dict)
        self.assertIn('RECORDS', adi_dict)
        self.assertEqual(0, len(adi_dict['HEADER']))
        self.assertEqual(2, len(adi_dict['RECORDS']))
        self.assertDictEqual(rec_dict, adi_dict['RECORDS'][0])

    def test_70_loadi(self):
        adi_txt = '''<QSO_DATE:8>20231008 <TIME_ON:4>1145 <CALL:6>dl4bdf<eor>
<QSO_DATE:8>20231008 <TIME_ON:4>1146 <CALL:6>DL5HJK <NAME:5>Peter <eor>
<QSO_DATE:8>20231008 <TIME_ON:4>1340 <RST_SENT:2>59<eor>
<QSO_DATE:8>20231008 <TIME_ON:4>1754<eor>
<eor>
<QSO_DATE:8>20231008 <MODE:2>AM <eor>'''

        rec_list = ({},
                    {'QSO_DATE': '20231008', 'TIME_ON': '1340', 'RST_SENT': '59'},
                    {'QSO_DATE': '20231008', 'TIME_ON': '1754'},
                    {},
                    {'QSO_DATE': '20231008', 'MODE': 'AM'})

        for exp, rec in zip(rec_list, adif_file.adi.loadi(adi_txt, 2)):
            self.assertDictEqual(exp, rec)

    def test_80_utf8file(self):
        adi_dict = adif_file.adi.load(get_file_path('testdata/utf8file.txt'), encoding='utf8')

        self.assertIn('HEADER', adi_dict)
        self.assertIn('RECORDS', adi_dict)
        self.assertEqual(3, len(adi_dict['HEADER']))
        self.assertEqual(5, len(adi_dict['RECORDS']))
        self.assertEqual('Jörg', adi_dict['RECORDS'][4]['NAME'])

    def test_81_latin1file(self):
        adi_dict = adif_file.adi.load(get_file_path('testdata/latin1file.txt'), encoding='latin1')

        self.assertIn('HEADER', adi_dict)
        self.assertIn('RECORDS', adi_dict)
        self.assertEqual(3, len(adi_dict['HEADER']))
        self.assertEqual(5, len(adi_dict['RECORDS']))
        self.assertEqual('Jörg', adi_dict['RECORDS'][4]['NAME'])


if __name__ == '__main__':
    unittest.main()



"""Convert ADIF ADX content to dictionary and vice versa
The XML is validated against the XSD from ADIF.org"""

import datetime
import os.path
import xml
from xml.etree.ElementTree import ElementTree, ParseError

import xmlschema
import xmltodict

from adif_file.__version__ import __version__ as __version_str__

__proj_name__ = 'PyADIF-File'
__version__ = __version_str__[1:].split('-')[0]

ADX_EXPORT_SCHEMA = xmlschema.XMLSchema(os.path.join(os.path.dirname(__file__), 'xsd/adx314.xsd'))
ADX_IMPORT_SCHEMA = xmlschema.XMLSchema(os.path.join(os.path.dirname(__file__), 'xsd/adx314generic.xsd'))


class MissingRecordsException(Exception):
    pass


class UndefinedElementException(Exception):
    pass


class MalformedValueException(Exception):
    pass


class XmlSyntaxError(SyntaxError):
    pass


def loads(adx_data: str, validate: bool = False) -> dict:
    """Load ADX content to dictionary
       The ADX is not validated to conform to the standard

       :param adx_data: the ADX content
       :param validate: validate the ADX against the genereic XSD (very slow)
       :return: the ADX as a dict
       """

    if validate:
        try:
            ADX_IMPORT_SCHEMA.validate(adx_data)
        except ParseError as exc:
            raise XmlSyntaxError(str(exc)) from None
        except xmlschema.validators.exceptions.XMLSchemaChildrenValidationError as exc:
            raise UndefinedElementException(f'in {exc.elem.tag}') from None
        except xmlschema.validators.exceptions.XMLSchemaValidationError as exc:
            raise MalformedValueException(f'Field "{exc.elem.tag}": {exc.reason}') from None

    try:
        data_dict = xmltodict.parse(adx_data, cdata_key='$')
        data_dict = data_dict['ADX']
        if ('RECORDS' in data_dict and data_dict['RECORDS'] and
                'RECORD' in data_dict['RECORDS'] and data_dict['RECORDS']['RECORD']):
            data_dict['RECORDS'] = data_dict['RECORDS']['RECORD']
        else:
            data_dict['RECORDS'] = []
        return data_dict
    except xml.parsers.expat.ExpatError as exc:
        raise XmlSyntaxError(str(exc)) from None


def load(file_name: str, validate: bool = False) -> dict:
    """Load ADX file to dictionary
       The XML is validated against the generic XSD

       :param file_name: the file name where the ADX data is stored
       :param validate: validate the ADX against the genereic XSD (very slow)
       :return: the ADX as a dict
       """

    with open(file_name, encoding='utf-8') as xf:
        adx_data = xf.read()

    return loads(adx_data, validate)


def dump(file_name: str, data_dict: dict):
    """Takes a dictionary and stores it to ADX xml file
       If 'HEADER' is missing the header fields are filled with defaults.
       The XML is validated against the strict XSD

       :param file_name: the filename to store the ADX data to
       :param data_dict: the dictionary with header and records
       """

    data_dict = data_dict.copy()

    header = {
        'ADIF_VER': '3.1.4',
        'PROGRAMID': __proj_name__,
        'PROGRAMVERSION': __version__,
        'CREATED_TIMESTAMP': datetime.datetime.utcnow().strftime('%Y%m%d %H%M%S')
    }

    if 'HEADER' not in data_dict:
        data_dict['HEADER'] = header
    else:
        for h in header:
            if h not in data_dict['HEADER']:
                data_dict['HEADER'][h] = header[h]

    if 'RECORDS' not in data_dict:
        raise MissingRecordsException('Missing records in data_dict')

    rec = data_dict.pop('RECORDS')
    data_dict['RECORDS'] = {'RECORD': rec}

    try:
        et = ADX_EXPORT_SCHEMA.encode(data_dict)
        ElementTree(et).write(file_name, xml_declaration=True, encoding='utf-8')
    except xmlschema.validators.exceptions.XMLSchemaChildrenValidationError as exc:
        raise UndefinedElementException(f'in {exc.elem.tag}') from None
    except xmlschema.validators.exceptions.XMLSchemaValidationError as exc:
        raise MalformedValueException(f'Field "{exc.elem.tag}": {exc.reason}') from None


__all__ = ['load', 'loads', 'dump', 'MissingRecordsException', 'UndefinedElementException',
           'MalformedValueException', 'XmlSyntaxError']



"""Convert ADIF ADI/ADX content to dictionary and vice versa"""

from adif_file.__version__ import __version__ as __version_str__

__author_name__ = 'Andreas Schawo, DF1ASC'
__author_email__ = 'andreas@schawo.de'
__copyright__ = 'Copyright 2023 by Andreas Schawo,licensed under CC BY-SA 4.0'
__proj_name__ = 'PyADIF-File'
__version__ = __version_str__[1:].split('-')[0]
__description__ = 'Convert ADIF ADI/ADX content to dictionary and vice versa'



import os
import unittest

import adif_file.adx


def get_file_path(file):
    return os.path.join(os.path.dirname(__file__), file)


class DumpADX(unittest.TestCase):
    def test_10_dump_exceptions(self):
        adx_dict = {
            'RECORDS': [{'CALL': 'XX1XXX',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1100',
                         'QTH': 'Test'}]
        }

        self.assertRaises(adif_file.adx.MissingRecordsException, adif_file.adx.dump, '', {})

        adx_dict2 = adx_dict.copy()
        adx_dict2['RECORDS'][0]['MY_QTH'] = 'Test'  # MY_QTH is not allowed (MY_CITY)
        self.assertRaises(adif_file.adx.UndefinedElementException, adif_file.adx.dump, '', adx_dict2)

        adx_dict2 = adx_dict.copy()
        adx_dict2['RECORDS'][0]['MY_CITY'] = 123
        self.assertRaises(adif_file.adx.MalformedValueException, adif_file.adx.dump, '', adx_dict2)

        adx_dict2 = adx_dict.copy()
        adx_dict2['RECORDS'][0]['QTH'] = 'Töst'  # QTH ony alows ASCII chars < 127
        self.assertRaises(adif_file.adx.MalformedValueException, adif_file.adx.dump, '', adx_dict2)

        adx_dict2 = adx_dict.copy()
        adx_dict2['RECORDS'][0]['FREQ'] = 'Test'
        self.assertRaises(adif_file.adx.MalformedValueException, adif_file.adx.dump, '', adx_dict2)

    def test_20_dump(self):
        adx_dict = {
            'HEADER': {'PROGRAMVERSION': '1',
                       'CREATED_TIMESTAMP': '20231204 100000',
                       },
            'RECORDS': [{'CALL': 'XX1XXX',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1100',
                         'QTH': 'Test'
                         },
                        {'CALL': 'YY1YYY',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1200',
                         'QTH_INTL': 'Töst',
                         'APP':
                             {
                                 '@PROGRAMID': 'TESTAPP',
                                 '@FIELDNAME': 'TESTFIELD',
                                 '@TYPE': 'I',
                                 '$': 'Test',
                             },
                         }]
        }

        adx_expected = '''<?xml version='1.0' encoding='utf-8'?>
<ADX>
    <HEADER>
        <PROGRAMVERSION>1</PROGRAMVERSION>
        <CREATED_TIMESTAMP>20231204 100000</CREATED_TIMESTAMP>
        <ADIF_VER>3.1.4</ADIF_VER>
        <PROGRAMID>PyADIF-File</PROGRAMID>
    </HEADER>
    <RECORDS>
        <RECORD>
            <CALL>XX1XXX</CALL>
            <QSO_DATE>20231204</QSO_DATE>
            <TIME_ON>1100</TIME_ON>
            <QTH>Test</QTH>
        </RECORD>
        <RECORD>
            <CALL>YY1YYY</CALL>
            <QSO_DATE>20231204</QSO_DATE>
            <TIME_ON>1200</TIME_ON>
            <QTH_INTL>Töst</QTH_INTL>
            <APP PROGRAMID="TESTAPP" FIELDNAME="TESTFIELD" TYPE="I">Test</APP>
        </RECORD>
    </RECORDS>
</ADX>
'''

        temp_file = get_file_path('testdata/~test.adx')
        adif_file.adx.dump(temp_file, adx_dict)

        with open(temp_file, encoding='utf-8') as af:
            self.assertEqual(adx_expected, af.read())

        os.remove(temp_file)

    def test_30_dump_no_change(self):
        adx_dict = {
            'HEADER': {'PROGRAMVERSION': '1',
                       'CREATED_TIMESTAMP': '20231204 100000',
                       },
            'RECORDS': [{'CALL': 'XX1XXX',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1100',
                         'QTH': 'Test'
                         },
                        {'CALL': 'YY1YYY',
                         'QSO_DATE': '20231204',
                         'TIME_ON': '1200',
                         'QTH_INTL': 'Töst',
                         'APP':
                             {
                                 '@PROGRAMID': 'TESTAPP',
                                 '@FIELDNAME': 'TESTFIELD',
                                 '@TYPE': 'I',
                                 '$': 'Test',
                             },
                         }]
        }
        adx_dict_sav = adx_dict.copy()

        adx_expected = '''<?xml version='1.0' encoding='utf-8'?>
<ADX>
    <HEADER>
        <PROGRAMVERSION>1</PROGRAMVERSION>
        <CREATED_TIMESTAMP>20231204 100000</CREATED_TIMESTAMP>
        <ADIF_VER>3.1.4</ADIF_VER>
        <PROGRAMID>PyADIF-File</PROGRAMID>
    </HEADER>
    <RECORDS>
        <RECORD>
            <CALL>XX1XXX</CALL>
            <QSO_DATE>20231204</QSO_DATE>
            <TIME_ON>1100</TIME_ON>
            <QTH>Test</QTH>
        </RECORD>
        <RECORD>
            <CALL>YY1YYY</CALL>
            <QSO_DATE>20231204</QSO_DATE>
            <TIME_ON>1200</TIME_ON>
            <QTH_INTL>Töst</QTH_INTL>
            <APP PROGRAMID="TESTAPP" FIELDNAME="TESTFIELD" TYPE="I">Test</APP>
        </RECORD>
    </RECORDS>
</ADX>
'''

        temp_file = get_file_path('testdata/~test.adx')
        adif_file.adx.dump(temp_file, adx_dict)
        self.assertDictEqual(adx_dict_sav, adx_dict)

        with open(temp_file, encoding='utf-8') as af:
            self.assertEqual(adx_expected, af.read())

        os.remove(temp_file)


if __name__ == '__main__':
    unittest.main()
